
from torch.utils.data import DataLoader, Dataset, Subset, random_split
import torch

# Custom Dataset
class TimeSeriesDataset(Dataset):
    def __init__(self, X, y, input_length=50, output_length=50, device="cpu"):
        self.X = torch.tensor(X, dtype=torch.float32).to(device)
        self.y = torch.tensor(y, dtype=torch.float32).to(device)
        self.input_length = input_length
        self.output_length = output_length

    def __len__(self):
        return len(self.X) - self.input_length - self.output_length

    def __getitem__(self, idx):
        x = self.X[idx:idx + self.input_length]
        y = self.y[idx + self.input_length:idx + self.input_length + self.output_length]
        return x, y


import numpy as np
import torch
import torch.nn as nn
class CumulLoss(nn.Module):
    def __init__(self, reduction='none'):
        super(CumulLoss, self).__init__()
        self.mae_loss = nn.L1Loss(reduction=reduction)

    def forward(self, outputs, labels):
        # Compute the cumulative sum along the sequence dimension (dim=1)
        cum_outputs = torch.cumsum(outputs, dim=1)
        cum_labels = torch.cumsum(labels, dim=1)

        # Compute the MAE loss between the cumulative sums of outputs and labels
        loss = self.mae_loss(cum_outputs, cum_labels)
        return loss



import torch
import torch.nn as nn
# Define the Model
class SimpleLinearModel(nn.Module):
    def __init__(self, input_size, output_size):
        super(SimpleLinearModel, self).__init__()
        self.layer1 = nn.Linear(input_size, 128)
        self.layer2 = nn.Linear(128, 128)
        self.layer3 = nn.Linear(128, output_size)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = x.view(x.size(0), -1)  # Flatten the input for the linear layers
        x = torch.relu(self.layer1(x))
        x = torch.relu(self.layer2(x))
        x = self.layer3(x)
        #x = self.sigmoid(x)
        return x.view(x.size(0), -1)  # Reshape the output back to (batch_size, output_length)


from .datasets import TimeSeriesDataset
from .models import SimpleLinearModel
from .utils import CumulLoss, plot_random_predictions
from .experiments import run_experiment  # Assuming `run_experiment` is a function in experiment_runner.py

__all__ = ["TimeSeriesDataset", "SimpleLinearModel", "CumulLoss", "plot_random_predictions", "run_experiment"]
__version__ = "0.1.1"




import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import pickle as pkl
import numpy as np
from matplotlib import pyplot as plt
from sklearn.preprocessing import StandardScaler
import random
import pickle as pkl
import torch
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, Dataset, Subset, random_split
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay, f1_score
import matplotlib.pyplot as plt
import os

def run_experiment(X, y , dataset, model, iterations=2, num_epochs=100, patience=10):
  f1_existence_list = []
  average_density_sum_list = []
  average_lead_time_list = []
  average_dice_score_list = []
  average_cumloss_scores_list = []

  train_loss_list = []
  val_loss_list = []
  test_loss_list = []

  # Use StratifiedKFold for 5-fold cross-validation
  #skf = StratifiedKFold(n_splits=5)
  skf = KFold(n_splits=5)

  input_length, output_length = dataset.input_length, dataset.input_length

  # Iterate through each fold
  for fold, (train_idx, test_idx) in enumerate(skf.split(X[input_length+output_length:, :], y[input_length+output_length:])):
      print(f'Fold {fold+1}')
      print(f'train_idx {train_idx}, test_idx {test_idx}, ')
      if fold != 4:
        continue
      
  for i in range(iterations):
      # Create train and test subsets
      train_dataset = Subset(dataset, train_idx)
      test_dataset = Subset(dataset, test_idx)

      # Calculate the validation set size
      val_size = len(train_dataset) // 8
      train_size = len(train_dataset) - val_size

      # Split train dataset into train and validation sets
      #train_subset, val_subset = random_split(train_dataset, [train_size, val_size])
      train_subset, val_subset = Subset(train_dataset, train_idx[:train_size]), Subset(train_dataset, train_idx[train_size:])

      train_loader = DataLoader(train_subset, batch_size=32, shuffle=True)
      val_loader = DataLoader(val_subset, batch_size=32, shuffle=False)
      test_loader = DataLoader(test_dataset, batch_size=32, shuffle=False)


      # Loss and optimizer
      criterion = CumulLoss(reduction='none')
      optimizer = optim.Adam(model.parameters(), lr=0.001)

      # Training loop with early stopping
      best_val_loss = float('inf')
      patience_counter = 0

      for epoch in range(num_epochs):
          train_loss = 0
          val_loss = 0
          test_loss = 0
          model.train()
          count = 0
          for inputs, labels in train_loader:
              optimizer.zero_grad()
              outputs = model(inputs)
              loss = criterion(outputs, labels).mean(1)
              loss.mean().backward()
              optimizer.step()
              train_loss += loss.sum().detach().item()
              count += len(inputs)
          train_loss /= count
          #print(f"train_loss, {train_loss:.03f}, {count}")

          model.eval()
          with torch.no_grad():
              count = 0
              for inputs, labels in val_loader:
                  outputs = model(inputs)
                  loss = criterion(outputs, labels).mean(1)
                  val_loss += loss.sum().item()
                  count += len(inputs)
              val_loss /= count
          #print(f"val_loss, {val_loss:.03f}, {count}")

          with torch.no_grad():
              count = 0
              for inputs, labels in test_loader:
                  outputs = model(inputs)
                  loss = criterion(outputs, labels).mean(1)
                  test_loss += loss.sum().item()
                  count += len(inputs)
              test_loss /= count
          #print(f"test_loss, {test_loss:.03f}, {count}")

          print(f'Epoch {epoch+1}/{num_epochs}, Training Loss: {train_loss}, Validation Loss: {val_loss}, Test Loss: {test_loss}')

          if val_loss < best_val_loss:
              best_val_loss = val_loss
              patience_counter = 0
              torch.save(model.state_dict(), f'best_model_fold{fold+1}.pth')  # Save the best model
          else:
              patience_counter += 1

          if patience_counter >= patience:
              print('Early stopping')
              break

          train_loss_list.append(train_loss)
          val_loss_list.append(val_loss)
          test_loss_list.append(test_loss)


      # Load the best model
      model.load_state_dict(torch.load(f'best_model_fold{fold+1}.pth'))

      # Metric 3: Dice Score
      def dice_score(true, pred):
          intersection = np.sum(true * pred)
          return (2. * intersection) / (np.sum(true) + np.sum(pred))

      # Evaluation on Test Set
      model.eval()
      existence_anomaly = {"TP": 0, "TN": 0, "FP": 0, "FN": 0}
      density_sum = []
      lead_time = []
      dice_scores = []
      cumloss_scores = []
      activation_threshold = 0.1

      with torch.no_grad():
          for inputs, labels in test_loader:
              outputs = model(inputs)
              loss = criterion(outputs, labels)
              outputs = outputs.detach().cpu().numpy()
              predictions = outputs > activation_threshold

              labels = labels.detach().cpu().numpy()

              exist_pred = predictions.sum(1) >= 1
              exist_labels = labels.sum(1) >= 1

              TP, FP, FN, TN = exist_pred * exist_labels, exist_pred * (1 - exist_labels), \
                              (1 - exist_pred) * exist_labels, (1 - exist_pred) * (1 - exist_labels)
              existence_anomaly["TP"] += TP.sum()
              existence_anomaly["TN"] += TN.sum()
              existence_anomaly["FP"] += FP.sum()
              existence_anomaly["FN"] += FN.sum()

              #density_sum.extend((1 - np.abs(outputs.mean(1) - labels.mean(1))).tolist())
              density_sum.extend([1 - abs(tf - pf) / len(outputs[1]) for tf, pf, true, pred in zip(labels, outputs, exist_labels, exist_pred) if true > 0 and pred > 0])
              true_first = np.argmax(labels, axis=1)
              pred_first = np.argmax(predictions, axis=1)

              lead_time.extend([1 - abs(tf - pf) / len(outputs[1]) for tf, pf, true, pred in zip(true_first, pred_first, exist_labels, exist_pred) if true > 0 and pred > 0])

              dice_scores.extend([dice_score(true, pred) for true, pred, true_exist, pred_exist  in zip(labels, predictions, exist_labels, exist_pred) if true_exist > 0 and pred_exist > 0])
              cumloss_scores.extend(loss.mean(1).tolist())

      # Print the result
      TP = existence_anomaly["TP"]
      TN = existence_anomaly["TN"]
      FP = existence_anomaly["FP"]
      FN = existence_anomaly["FN"]

      precision = TP / (TP + FP) if (TP + FP) > 0 else 0
      recall = TP / (TP + FN) if (TP + FN) > 0 else 0
      f1_existence = 2 * ( TP ) / (2 * TP + FP + FN) if (2 * TP + FP + FN) > 0 else 0

      average_density_sum = np.mean(density_sum)
      average_lead_time = np.mean(lead_time) if lead_time else float('nan')
      average_dice_score = np.mean(dice_scores)
      average_cumloss_scores = np.mean(cumloss_scores) / (dataset.output_length+1) * 2 # coeff 1/(T*(T+1))

      f1_existence_list.append(f1_existence)
      average_density_sum_list.append(average_density_sum)
      average_lead_time_list.append(average_lead_time)
      average_dice_score_list.append(average_dice_score)
      average_cumloss_scores_list.append(average_cumloss_scores)

  return TP, TN, FP, FN, f1_existence, \
            average_density_sum_list, average_lead_time_list, \
            average_dice_score_list, average_cumloss_scores_list, \
                train_loss_list, val_loss_list, test_loss_list, \
                train_loader, val_loader, test_loader


import matplotlib.pyplot as plt
import torch
import numpy as np

def plot_random_predictions(dataloader, model, num_predictions=3):
    """
    Plot randomly selected predictions and ground truth values from the dataloader.

    Parameters:
    - dataloader (DataLoader): The DataLoader to get data from.
    - model (nn.Module): The trained model used for making predictions.
    - num_predictions (int): The number of random predictions to plot (default is 3).
    """
    model.eval()
    sample_list = []
    for (inputs, labels) in dataloader:
        for i in range(len(inputs)):
            if labels[i].sum() > 1.0 and np.random.random() < 0.2:
                sample_list.append((inputs[i], labels[i]))
            if len(sample_list) > num_predictions:
                break

    fig, axes = plt.subplots(1, num_predictions, figsize=(5*num_predictions, num_predictions))

    activation_threshold = 0.1
    with torch.no_grad():
        for i, (inputs, labels) in enumerate(sample_list):
            if i >= num_predictions:
                break
            inputs = torch.unsqueeze(inputs.float(), 0)
            outputs = model(inputs).detach().cpu().numpy()
            inputs = inputs[0].detach().cpu().numpy()
            labels = labels.detach().cpu().numpy()
            predictions = outputs > activation_threshold
            
            # Plot input series on the primary y-axis
            for k in range(inputs.shape[1]):
                axes[i].plot(inputs[:, k].flatten(), label=f'Input Series ({k})', alpha=0.7)
            
            # Create a secondary y-axis
            ax2 = axes[i].twinx()
            ax2.plot(np.arange(len(inputs[:, 0].flatten()), len(inputs[:, 0].flatten()) + len(labels.flatten())), labels.flatten(), label='Ground Truth', color='green')
            ax2.plot(np.arange(len(inputs[:, 0].flatten()), len(inputs[:, 0].flatten()) + len(predictions.flatten())), predictions.flatten(), label='Prediction', color='red', linestyle='-', alpha=0.3)
            ax2.plot(np.arange(len(inputs[:, 0].flatten()), len(inputs[:, 0].flatten()) + len(outputs.flatten())), outputs.flatten(), label='Density', color='orange', linestyle='dashed')
            
            # Set the y-axis limit of the secondary axis to [0, 1]
            ax2.set_ylim(-0.1, 1.1)
            
            # Plot the vertical line on both axes
            axes[i].axvline(x=len(inputs[:, 0].flatten()) - 1, color='black', linestyle='--')
            ax2.axvline(x=len(inputs[:, 0].flatten()) - 1, color='black', linestyle='--')
            
            # Set legends and titles
            axes[i].legend(loc='upper left')
            ax2.legend(loc='lower left')
            axes[i].set_title(f'Prediction {i+1}')

    plt.tight_layout()
    plt.show()

def plot_random_predictions_heatmap(dataloader, model, num_predictions=3):
    """
    Plot randomly selected predictions and ground truth values from the dataloader.

    Parameters:
    - dataloader (DataLoader): The DataLoader to get data from.
    - model (nn.Module): The trained model used for making predictions.
    - num_predictions (int): The number of random predictions to plot (default is 3).
    """
    model.eval()
    sample_list = []
    for (inputs, labels) in dataloader:
        for i in range(len(inputs)):
            if labels[i].sum() > 1.0 and np.random.random() < 0.2:
                sample_list.append((inputs[i], labels[i]))
            if len(sample_list) > num_predictions:
                break

    fig, axes = plt.subplots(1, num_predictions, figsize=(5*num_predictions, num_predictions))

    # Ensure axes is iterable
    if num_predictions == 1:
        axes = [axes]

    with torch.no_grad():
        for i, (inputs, labels) in enumerate(sample_list):
            if i >= num_predictions:
                break
            inputs = torch.unsqueeze(inputs.float(), 0).to(device)
            outputs = model(inputs).detach().cpu().numpy()
            inputs = inputs[0].detach().cpu().numpy()
            labels = labels.detach().cpu().numpy()
            predictions = outputs > 0.1

            # Plot the inputs as heatmaps
            im = axes[i].imshow(inputs.T, aspect='auto', origin='lower', interpolation='nearest', cmap='viridis')

            # Add colorbar
            # fig.colorbar(im, ax=axes[i])

            # Create a secondary y-axis for outputs
            ax2 = axes[i].twinx()

            # Adjust x-axis for labels and predictions
            input_length = inputs.shape[0]
            time_labels = np.arange(input_length, input_length + len(labels.flatten()))

            # Plot the outputs on the secondary y-axis
            ax2.plot(time_labels, labels.flatten(), label='Ground Truth', color='green')
            ax2.plot(time_labels, predictions.flatten(), label='Prediction', color='red', linestyle='-', alpha=0.3)
            ax2.plot(time_labels, outputs.flatten(), label='Density', color='orange', linestyle='dashed')

            # Plot the vertical line
            axes[i].axvline(x=input_length - 1, color='black', linestyle='--')

            # Hide the y-axis scale for the input (primary axis)
            axes[i].tick_params(axis='y', which='both', left=False, labelleft=False)
            axes[i].set_ylabel('')  # Remove y-axis label

            # Set y-axis label for the outputs (secondary axis)
            ax2.set_ylabel('Output Values')

            # Create custom legend entries
            handles1, labels1 = axes[i].get_legend_handles_labels()
            handles2, labels2 = ax2.get_legend_handles_labels()

            # Combine handles and labels
            handles = handles1 + handles2
            legend_labels = labels1 + labels2

            # Create a custom patch for the input heatmap
            cmap = plt.get_cmap('viridis')
            midpoint_color = cmap(0.5)
            input_patch = Patch(color=midpoint_color, label='Input Heatmap')

            # Insert the input_patch at the beginning of the handles and labels
            handles.insert(0, input_patch)
            legend_labels.insert(0, 'Input Heatmap')

            # Update the legend with the custom handles and labels
            axes[i].legend(handles=handles, labels=legend_labels, loc='upper left')

            axes[i].set_title(f'Prediction {i+1}')

            # Optionally, set the x-axis label
            axes[i].set_xlabel('Time')

        plt.tight_layout()
        plt.show()


from setuptools import setup, find_packages

setup(
    name="anomaly-prediction",                  # Package name (used for pip install)
    version="0.1.1",                            # Initial version
    description="A library for anomaly prediction in time series data.",
    long_description=open("readme.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Jiang YOU",
    author_email="jiang.you@esiee.fr",
    url="https://github.com/JiangYou2025/AnomalyPrediction",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[                          # List of dependencies
        "numpy",
        "torch",
        "matplotlib",
        "sklearn",
        # Add any other dependencies your library needs
    ],
    classifiers=[                               # Additional metadata
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)
